@@ -1,9 +1,7 @@
 use std::alloc::{alloc, dealloc, handle_alloc_error, Layout};
-use std::any::{type_name, TypeId};
-use std::cell::UnsafeCell;
+use std::any::TypeId;
 use std::cmp::Reverse;
-use std::mem::align_of;
-use std::ops::{Deref, DerefMut};
+use std::mem::{align_of, needs_drop};
 use std::ptr::{copy_nonoverlapping, drop_in_place};
 use std::slice::from_raw_parts_mut;
 
@@ -233,26 +231,6 @@ impl Archetype {
         ty
     }
 
-    pub unsafe fn borrow<C>(&self, ty: u16) -> Ref<'_>
-    where
-        C: 'static,
-    {
-        let ty = self.type_metadata::<C>(ty);
-
-        Ref::new(&ty.borrow)
-            .unwrap_or_else(|| panic!("Component {} already borrowed", type_name::<C>()))
-    }
-
-    pub unsafe fn borrow_mut<C>(&self, ty: u16) -> RefMut<'_>
-    where
-        C: 'static,
-    {
-        let ty = self.type_metadata::<C>(ty);
-
-        RefMut::new(&ty.borrow)
-            .unwrap_or_else(|| panic!("Component {} already borrowed", type_name::<C>()))
-    }
-
     pub unsafe fn base_pointer<C>(&self, ty: u16) -> *mut C
     where
         C: 'static,
@@ -275,7 +253,7 @@ impl Archetype {
 }
 
 impl Archetype {
-    pub unsafe fn get_raw<C>(&mut self, idx: u32) -> *mut C
+    pub unsafe fn get<C>(&self, idx: u32) -> *mut C
     where
         C: 'static,
     {
@@ -283,141 +261,28 @@ impl Archetype {
         self.pointer::<C>(ty, idx)
     }
 
-    pub unsafe fn get<C>(&self, idx: u32) -> Option<Comp<'_, C>>
+    pub unsafe fn drop<C>(&mut self, idx: u32)
     where
         C: 'static,
     {
-        let ty = self.find::<C>()?;
-        let _ref = self.borrow::<C>(ty);
-        let ptr = self.pointer::<C>(ty, idx);
+        if needs_drop::<C>() {
+            if let Some(ty) = self.find::<C>() {
+                let ty = self.types.get_unchecked(ty as usize);
 
-        let val = &*ptr;
+                let ptr = ty.base_pointer.add(ty.layout.size() * idx as usize);
 
-        Some(Comp { _ref, val })
-    }
-
-    pub unsafe fn get_mut<C>(&self, idx: u32) -> Option<CompMut<'_, C>>
-    where
-        C: 'static,
-    {
-        let ty = self.find::<C>()?;
-        let _ref = self.borrow_mut::<C>(ty);
-        let ptr = self.pointer::<C>(ty, idx);
-
-        let val = &mut *ptr;
-
-        Some(CompMut { _ref, val })
-    }
-}
-
-/// An immutable borrow of a component.
-pub struct Comp<'a, C> {
-    _ref: Ref<'a>,
-    val: &'a C,
-}
-
-impl<C> Deref for Comp<'_, C> {
-    type Target = C;
-
-    fn deref(&self) -> &C {
-        self.val
-    }
-}
-
-/// A mutable borrow of a component.
-pub struct CompMut<'a, C> {
-    _ref: RefMut<'a>,
-    val: &'a mut C,
-}
-
-impl<C> Deref for CompMut<'_, C> {
-    type Target = C;
-
-    fn deref(&self) -> &C {
-        self.val
-    }
-}
-
-impl<C> DerefMut for CompMut<'_, C> {
-    fn deref_mut(&mut self) -> &mut C {
-        self.val
-    }
-}
-
-pub struct Ref<'a>(&'a UnsafeCell<isize>);
-
-impl<'a> Ref<'a> {
-    fn new(borrow: &'a UnsafeCell<isize>) -> Option<Self> {
-        let readers = unsafe { &mut *borrow.get() };
-
-        let new_readers = readers.wrapping_add(1);
-
-        if new_readers <= 0 {
-            cold();
-            return None;
-        }
-
-        *readers = new_readers;
-
-        Some(Self(borrow))
-    }
-}
-
-impl Drop for Ref<'_> {
-    fn drop(&mut self) {
-        unsafe {
-            *self.0.get() -= 1;
+                (ty.drop)(ptr, 1);
+            }
         }
     }
 }
 
-pub struct RefMut<'a>(&'a UnsafeCell<isize>);
-
-impl<'a> RefMut<'a> {
-    fn new(borrow: &'a UnsafeCell<isize>) -> Option<Self> {
-        let writers = unsafe { &mut *borrow.get() };
-
-        if *writers != 0 {
-            cold();
-            return None;
-        }
-
-        *writers = -1;
-
-        Some(Self(borrow))
-    }
-}
-
-impl Drop for RefMut<'_> {
-    fn drop(&mut self) {
-        unsafe {
-            *self.0.get() = 0;
-        }
-    }
-}
-
-#[cold]
-#[inline(always)]
-fn cold() {}
-
+#[derive(Clone, Copy)]
 struct TypeMetadata {
     id: TypeId,
     layout: Layout,
     drop: unsafe fn(*mut u8, usize),
     base_pointer: *mut u8,
-    borrow: UnsafeCell<isize>,
-}
-
-impl Clone for TypeMetadata {
-    fn clone(&self) -> Self {
-        Self {
-            id: self.id,
-            layout: self.layout,
-            drop: self.drop,
-            base_pointer: self.layout.align() as *mut u8,
-            borrow: Default::default(),
-        }
-    }
 }
 
 impl TypeMetadata {
@@ -434,7 +299,6 @@ impl TypeMetadata {
             layout: Layout::new::<C>(),
             drop: drop::<C>,
             base_pointer: align_of::<C>() as *mut u8,
-            borrow: Default::default(),
         }
     }
 }
@@ -443,15 +307,16 @@ impl TypeMetadata {
 pub struct TypeMetadataSet(Vec<TypeMetadata>);
 
 impl TypeMetadataSet {
+    pub fn ids(&self) -> impl ExactSizeIterator<Item = TypeId> + '_ {
+        self.0.iter().map(|ty| ty.id)
+    }
+
     pub fn insert<C>(&mut self)
     where
         C: 'static,
     {
-        let ty = self.0.binary_search_by_key(&TypeId::of::<C>(), |ty| ty.id);
-
-        match ty {
-            Err(ty) => self.0.insert(ty, TypeMetadata::new::<C>()),
-            Ok(_) => panic!("Component {} already present", type_name::<C>()),
+        if let Err(ty) = self.0.binary_search_by_key(&TypeId::of::<C>(), |ty| ty.id) {
+            self.0.insert(ty, TypeMetadata::new::<C>());
         }
     }
 
@@ -475,7 +340,8 @@ impl TypeMetadataSet {
 mod tests {
     use super::*;
 
-    use std::sync::atomic::{AtomicUsize, Ordering};
+    use std::cell::Cell;
+    use std::rc::Rc;
 
     #[test]
     fn reverse_sorting_by_alignment_avoids_padding() {
@@ -517,16 +383,18 @@ mod tests {
 
     #[test]
     fn drops_all_component_values() {
-        static DROPS: AtomicUsize = AtomicUsize::new(0);
-
-        struct CountDrops;
+        struct CountDrops(Rc<Cell<usize>>);
 
         impl Drop for CountDrops {
             fn drop(&mut self) {
-                DROPS.fetch_add(1, Ordering::Relaxed);
+                let drops = &self.0;
+
+                drops.set(drops.get() + 1);
             }
         }
 
+        let drops = Rc::new(Cell::new(0));
+
         let mut types = TypeMetadataSet::default();
         types.insert::<CountDrops>();
 
@@ -536,12 +404,14 @@ mod tests {
             for _ in 0..32 {
                 let ent = archetype.alloc();
 
-                archetype.get_raw::<CountDrops>(ent).write(CountDrops);
+                archetype
+                    .get::<CountDrops>(ent)
+                    .write(CountDrops(drops.clone()));
             }
         }
 
         drop(archetype);
 
-        assert_eq!(DROPS.load(Ordering::Relaxed), 32);
+        assert_eq!(drops.get(), 32);
     }
 }
