@@ -3,15 +3,20 @@ use std::collections::HashMap;
 use std::convert::TryInto;
 use std::hash::{BuildHasherDefault, Hasher};
 use std::num::NonZeroU32;
+use std::ops::{Deref, DerefMut};
 use std::sync::atomic::{AtomicU32, Ordering};
 
-use crate::archetype::{Archetype, Comp, CompMut, TypeMetadataSet};
+use crate::{
+    archetype::{Archetype, TypeMetadataSet},
+    borrow_flags::{BorrowFlags, Ref, RefMut},
+};
 
 /// The world storing entities and their components.
 pub struct World {
     tag: u32,
     pub(crate) entities: Vec<EntityMetadata>,
     free_list: Vec<u32>,
+    pub(crate) borrow_flags: BorrowFlags,
     pub(crate) archetypes: Vec<Archetype>,
     insert_map: IndexTypeIdMap<u16>,
     remove_map: IndexTypeIdMap<u16>,
@@ -32,11 +37,17 @@ impl World {
         let mut empty_archetype = TypeMetadataSet::default();
         empty_archetype.insert::<Entity>();
 
+        let mut borrow_flags = BorrowFlags::default();
+        borrow_flags.insert(&empty_archetype);
+
+        let archetypes = vec![Archetype::new(empty_archetype)];
+
         Self {
             tag: tag(),
             entities: Default::default(),
             free_list: Default::default(),
-            archetypes: vec![Archetype::new(empty_archetype)],
+            borrow_flags,
+            archetypes,
             insert_map: Default::default(),
             remove_map: Default::default(),
             exchange_map: Default::default(),
@@ -79,7 +90,7 @@ impl World {
         let ent = Entity { id, gen: meta.gen };
 
         unsafe {
-            self.archetypes[0].get_raw::<Entity>(meta.idx).write(ent);
+            self.archetypes[0].get::<Entity>(meta.idx).write(ent);
         }
 
         ent
@@ -131,7 +142,7 @@ impl World {
     ) {
         unsafe {
             if archetype.free::<DROP>(idx) {
-                let swapped_ent = archetype.get_raw::<Entity>(idx).read();
+                let swapped_ent = archetype.get::<Entity>(idx).read();
 
                 entities[swapped_ent.id as usize].idx = idx;
             }
@@ -157,9 +168,7 @@ impl World {
 
     /// Insert components for a given [Entity].
     ///
-    /// # Panics
-    ///
-    /// Panics if one of the components is already present for the entity.
+    /// If a component is already present for the entity, its value will be overwritten.
     ///
     /// # Example
     ///
@@ -185,8 +194,8 @@ impl World {
         } else {
             Self::insert_cold(
                 &mut self.archetypes,
+                &mut self.borrow_flags,
                 &mut self.insert_map,
-                &mut self.remove_map,
                 key,
                 B::insert,
                 meta.ty,
@@ -197,6 +206,8 @@ impl World {
         let old_idx = meta.idx;
 
         unsafe {
+            B::drop::<()>(&mut self.archetypes[old_ty as usize], old_idx);
+
             let new_idx = self.move_(ent.id, old_ty, new_ty, old_idx);
 
             comps.write(&mut self.archetypes[new_ty as usize], new_idx);
@@ -207,8 +218,8 @@ impl World {
     #[inline(never)]
     fn insert_cold(
         archetypes: &mut Vec<Archetype>,
+        borrow_flags: &mut BorrowFlags,
         insert_map: &mut IndexTypeIdMap<u16>,
-        remove_map: &mut IndexTypeIdMap<u16>,
         key: TypeId,
         insert: fn(&mut TypeMetadataSet),
         old_ty: u16,
@@ -216,10 +227,9 @@ impl World {
         let mut types = archetypes[old_ty as usize].types();
         insert(&mut types);
 
-        let new_ty = Self::get_or_insert(archetypes, types);
+        let new_ty = Self::get_or_insert(archetypes, borrow_flags, types);
 
         insert_map.insert((old_ty, key), new_ty);
-        remove_map.insert((new_ty, key), old_ty);
 
         new_ty
     }
@@ -252,7 +262,7 @@ impl World {
         } else {
             Self::remove_cold(
                 &mut self.archetypes,
-                &mut self.insert_map,
+                &mut self.borrow_flags,
                 &mut self.remove_map,
                 key,
                 B::remove,
@@ -276,7 +286,7 @@ impl World {
     #[inline(never)]
     fn remove_cold(
         archetypes: &mut Vec<Archetype>,
-        insert_map: &mut IndexTypeIdMap<u16>,
+        borrow_flags: &mut BorrowFlags,
         remove_map: &mut IndexTypeIdMap<u16>,
         key: TypeId,
         remove: fn(&mut TypeMetadataSet) -> Option<()>,
@@ -285,10 +295,9 @@ impl World {
         let mut types = archetypes[old_ty as usize].types();
         remove(&mut types)?;
 
-        let new_ty = Self::get_or_insert(archetypes, types);
+        let new_ty = Self::get_or_insert(archetypes, borrow_flags, types);
 
         remove_map.insert((old_ty, key), new_ty);
-        insert_map.insert((new_ty, key), old_ty);
 
         Some(new_ty)
     }
@@ -327,6 +336,7 @@ impl World {
         } else {
             Self::exchange_cold(
                 &mut self.archetypes,
+                &mut self.borrow_flags,
                 &mut self.exchange_map,
                 key,
                 B1::remove,
@@ -339,13 +349,13 @@ impl World {
         let old_idx = meta.idx;
 
         unsafe {
-            let old_comps = B1::read(&mut self.archetypes[old_ty as usize], old_idx);
+            let old_archetype = &mut self.archetypes[old_ty as usize];
 
-            let new_idx = if new_ty != old_ty {
-                self.move_(ent.id, old_ty, new_ty, old_idx)
-            } else {
-                old_idx
-            };
+            let old_comps = B1::read(old_archetype, old_idx);
+
+            B2::drop::<B1>(old_archetype, old_idx);
+
+            let new_idx = self.move_(ent.id, old_ty, new_ty, old_idx);
 
             new_comps.write(&mut self.archetypes[new_ty as usize], new_idx);
 
@@ -357,6 +367,7 @@ impl World {
     #[inline(never)]
     fn exchange_cold(
         archetypes: &mut Vec<Archetype>,
+        borrow_flags: &mut BorrowFlags,
         exchange_map: &mut IndexTypeIdMap<u16>,
         key: TypeId,
         remove: fn(&mut TypeMetadataSet) -> Option<()>,
@@ -367,14 +378,18 @@ impl World {
         remove(&mut types)?;
         insert(&mut types);
 
-        let new_ty = Self::get_or_insert(archetypes, types);
+        let new_ty = Self::get_or_insert(archetypes, borrow_flags, types);
 
         exchange_map.insert((old_ty, key), new_ty);
 
         Some(new_ty)
     }
 
-    fn get_or_insert(archetypes: &mut Vec<Archetype>, types: TypeMetadataSet) -> u16 {
+    fn get_or_insert(
+        archetypes: &mut Vec<Archetype>,
+        borrow_flags: &mut BorrowFlags,
+        types: TypeMetadataSet,
+    ) -> u16 {
         let pos = archetypes
             .iter_mut()
             .position(|archetype| archetype.match_(&types));
@@ -385,6 +400,8 @@ impl World {
             let len = archetypes.len();
             assert!(len < u16::MAX as usize);
 
+            borrow_flags.insert(&types);
+
             archetypes.push(Archetype::new(types));
 
             len as u16
@@ -392,9 +409,12 @@ impl World {
     }
 
     unsafe fn move_(&mut self, id: u32, old_ty: u16, new_ty: u16, old_idx: u32) -> u32 {
+        if old_ty == new_ty {
+            return old_idx;
+        }
+
         debug_assert!(self.archetypes.len() > old_ty as usize);
         debug_assert!(self.archetypes.len() > new_ty as usize);
-        debug_assert_ne!(old_ty, new_ty);
 
         let archetypes = self.archetypes.as_mut_ptr();
         let old_archetype = &mut *archetypes.add(old_ty as usize);
@@ -453,6 +473,7 @@ impl World {
             Self::transfer_cold(
                 &self.archetypes,
                 &mut other.archetypes,
+                &mut other.borrow_flags,
                 &mut self.transfer_map,
                 &mut other.transfer_map,
                 self.tag,
@@ -481,18 +502,20 @@ impl World {
 
         unsafe {
             other.archetypes[new_meta.ty as usize]
-                .get_raw::<Entity>(new_meta.idx)
+                .get::<Entity>(new_meta.idx)
                 .write(ent);
         }
 
         ent
     }
 
+    #[allow(clippy::too_many_arguments)]
     #[cold]
     #[inline(never)]
     fn transfer_cold(
         archetypes: &[Archetype],
         other_archetypes: &mut Vec<Archetype>,
+        other_borrows: &mut BorrowFlags,
         transfer_map: &mut IndexTagMap<u16>,
         other_transfer_map: &mut IndexTagMap<u16>,
         tag: u32,
@@ -501,7 +524,7 @@ impl World {
     ) -> u16 {
         let types = archetypes[old_ty as usize].types();
 
-        let new_ty = Self::get_or_insert(other_archetypes, types);
+        let new_ty = Self::get_or_insert(other_archetypes, other_borrows, types);
 
         transfer_map.insert((old_ty, other_tag), new_ty);
         other_transfer_map.insert((new_ty, tag), old_ty);
@@ -575,7 +598,15 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        unsafe { self.archetypes[meta.ty as usize].get::<C>(meta.idx) }
+        let archetype = &self.archetypes[meta.ty as usize];
+        let comp = archetype.find::<C>()?;
+
+        let flag = self.borrow_flags.find::<C>()?;
+        let _ref = unsafe { self.borrow_flags.borrow::<C>(flag) };
+
+        let val = unsafe { &*archetype.pointer::<C>(comp, meta.idx) };
+
+        Some(Comp { _ref, val })
     }
 
     /// Get a mutable reference to the component of the given type for an [Entity].
@@ -606,35 +637,15 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        unsafe { self.archetypes[meta.ty as usize].get_mut::<C>(meta.idx) }
-    }
+        let archetype = &self.archetypes[meta.ty as usize];
+        let comp = archetype.find::<C>()?;
 
-    /// Either update or insert the value of component `C` at entity `ent`.
-    ///
-    /// # Example
-    ///
-    /// ```
-    /// # use rs_ecs::*;
-    /// let mut world = World::new();
-    ///
-    /// let entity1 = world.alloc();
-    /// world.insert(entity1, (0,));
-    ///
-    /// let entity2 = world.alloc();
-    ///
-    /// world.upsert(entity1, 1);
-    /// world.upsert(entity2, 2);
-    /// ```
-    pub fn upsert<C>(&mut self, ent: Entity, new_comp: C)
-    where
-        C: 'static,
-    {
-        if let Some(mut comp) = self.get_mut::<C>(ent) {
-            *comp = new_comp;
-            return;
-        }
+        let flag = self.borrow_flags.find::<C>()?;
+        let _ref = unsafe { self.borrow_flags.borrow_mut::<C>(flag) };
 
-        self.insert(ent, (new_comp,));
+        let val = unsafe { &mut *archetype.pointer::<C>(comp, meta.idx) };
+
+        Some(CompMut { _ref, val })
     }
 }
 
@@ -668,17 +679,65 @@ impl EntityMetadata {
     }
 }
 
+/// An immutable borrow of a component.
+pub struct Comp<'a, C> {
+    _ref: Ref<'a>,
+    val: &'a C,
+}
+
+impl<C> Deref for Comp<'_, C> {
+    type Target = C;
+
+    fn deref(&self) -> &C {
+        self.val
+    }
+}
+
+/// A mutable borrow of a component.
+pub struct CompMut<'a, C> {
+    _ref: RefMut<'a>,
+    val: &'a mut C,
+}
+
+impl<C> Deref for CompMut<'_, C> {
+    type Target = C;
+
+    fn deref(&self) -> &C {
+        self.val
+    }
+}
+
+impl<C> DerefMut for CompMut<'_, C> {
+    fn deref_mut(&mut self) -> &mut C {
+        self.val
+    }
+}
+
 #[allow(clippy::missing_safety_doc)]
-pub unsafe trait Bundle: 'static {
+pub unsafe trait Bundle
+where
+    Self: 'static,
+{
+    fn contains<T>() -> bool
+    where
+        T: 'static;
+
     fn insert(types: &mut TypeMetadataSet);
     #[must_use]
     fn remove(types: &mut TypeMetadataSet) -> Option<()>;
+
+    unsafe fn drop<T>(archetype: &mut Archetype, idx: u32)
+    where
+        T: Bundle;
+
     unsafe fn write(self, archetype: &mut Archetype, idx: u32);
     unsafe fn read(archetype: &mut Archetype, idx: u32) -> Self;
 }
 
 macro_rules! impl_bundle_for_tuples {
-    () => {};
+    () => {
+        impl_bundle_for_tuples!(@impl);
+    };
 
     ($head:ident $(,$tail:ident)*) => {
         impl_bundle_for_tuples!($($tail),*);
@@ -693,13 +752,35 @@ macro_rules! impl_bundle_for_tuples {
         impl_bundle_for_tuples!(@rev $($tail),*; $head $(,$rev)*);
     };
 
-    (@impl $($types:ident),+) => {
-        unsafe impl<$($types),+> Bundle for ($($types,)+)
+    (@impl $($types:ident),*) => {
+        #[allow(unused_variables)]
+        unsafe impl<$($types),*> Bundle for ($($types,)*)
         where
-            $($types: 'static,)+
+            $($types: 'static,)*
         {
+            fn contains<T>() -> bool
+            where
+                T: 'static
+            {
+                $(
+                    if TypeId::of::<$types>() == TypeId::of::<T>() {
+                        return true;
+                    }
+                )*
+
+                false
+            }
+
             fn insert(types: &mut TypeMetadataSet) {
-                $(types.insert::<$types>();)+
+                $(
+                    assert_ne!(
+                        TypeId::of::<$types>(),
+                        TypeId::of::<Entity>(),
+                        "Entity cannot be inserted"
+                    );
+
+                    types.insert::<$types>();
+                )*
             }
 
             fn remove(types: &mut TypeMetadataSet) -> Option<()> {
@@ -711,21 +792,33 @@ macro_rules! impl_bundle_for_tuples {
                     );
 
                     types.remove::<$types>()?;
-                )+
+                )*
 
                 Some(())
             }
 
-            #[allow(non_snake_case)]
-            unsafe fn write(self, archetype: &mut Archetype, idx: u32) {
-                let ($($types,)+) = self;
-                $(archetype.get_raw::<$types>(idx).write($types);)+
+            unsafe fn drop<T>(archetype: &mut Archetype, idx: u32)
+            where
+                T: Bundle
+            {
+                $(
+                    if !T::contains::<$types>() {
+                        archetype.drop::<$types>(idx);
+                    }
+                )*
             }
 
             #[allow(non_snake_case)]
+            unsafe fn write(self, archetype: &mut Archetype, idx: u32) {
+                let ($($types,)*) = self;
+                $(archetype.get::<$types>(idx).write($types);)*
+            }
+
+            #[allow(non_snake_case)]
+            #[allow(clippy::unused_unit)]
             unsafe fn read(archetype: &mut Archetype, idx: u32) -> Self {
-                $(let $types = archetype.get_raw::<$types>(idx).read();)+
-                ($($types,)+)
+                $(let $types = archetype.get::<$types>(idx).read();)*
+                ($($types,)*)
             }
         }
     };
@@ -783,9 +876,19 @@ impl Hasher for IndexTagHasher {
 mod tests {
     use super::*;
 
+    use std::cell::Cell;
     #[cfg(not(miri))]
     use std::hash::Hash;
     use std::mem::size_of;
+    use std::rc::Rc;
+
+    struct SetOnDrop(Rc<Cell<bool>>);
+
+    impl Drop for SetOnDrop {
+        fn drop(&mut self) {
+            self.0.set(true);
+        }
+    }
 
     #[test]
     fn alloc_creates_unique_entities() {
@@ -853,7 +956,6 @@ mod tests {
         assert_eq!(world.archetypes[0].len(), 0);
 
         assert_eq!(world.insert_map.len(), 0);
-        assert_eq!(world.remove_map.len(), 0);
 
         let ent = world.alloc();
 
@@ -866,7 +968,6 @@ mod tests {
         assert_eq!(world.archetypes[0].len(), 1);
 
         assert_eq!(world.insert_map.len(), 0);
-        assert_eq!(world.remove_map.len(), 0);
 
         world.insert(ent, (23_i32,));
 
@@ -881,9 +982,6 @@ mod tests {
 
         assert_eq!(world.insert_map.len(), 1);
         assert_eq!(world.insert_map[&(0, TypeId::of::<(i32,)>())], 1);
-
-        assert_eq!(world.remove_map.len(), 1);
-        assert_eq!(world.remove_map[&(1, TypeId::of::<(i32,)>())], 0);
     }
 
     #[test]
@@ -900,17 +998,6 @@ mod tests {
 
         let ent = world.alloc();
 
-        assert_eq!(world.entities.len(), 1);
-        assert_eq!(world.entities[0].gen.get(), 1);
-        assert_eq!(world.entities[0].ty, 0);
-        assert_eq!(world.entities[0].idx, 0);
-
-        assert_eq!(world.archetypes.len(), 1);
-        assert_eq!(world.archetypes[0].len(), 1);
-
-        assert_eq!(world.insert_map.len(), 0);
-        assert_eq!(world.remove_map.len(), 0);
-
         world.insert(ent, (23_i32, 42_u64));
 
         assert_eq!(world.entities.len(), 1);
@@ -922,11 +1009,7 @@ mod tests {
         assert_eq!(world.archetypes[0].len(), 0);
         assert_eq!(world.archetypes[1].len(), 1);
 
-        assert_eq!(world.insert_map.len(), 1);
-        assert_eq!(world.insert_map[&(0, TypeId::of::<(i32, u64)>())], 1);
-
-        assert_eq!(world.remove_map.len(), 1);
-        assert_eq!(world.remove_map[&(1, TypeId::of::<(i32, u64)>())], 0);
+        assert_eq!(world.remove_map.len(), 0);
 
         world.remove::<(i32,)>(ent).unwrap();
 
@@ -940,15 +1023,74 @@ mod tests {
         assert_eq!(world.archetypes[1].len(), 0);
         assert_eq!(world.archetypes[2].len(), 1);
 
-        assert_eq!(world.insert_map.len(), 2);
-        assert_eq!(world.insert_map[&(0, TypeId::of::<(i32, u64)>())], 1);
-        assert_eq!(world.insert_map[&(2, TypeId::of::<(i32,)>())], 1);
-
-        assert_eq!(world.remove_map.len(), 2);
-        assert_eq!(world.remove_map[&(1, TypeId::of::<(i32, u64)>())], 0);
+        assert_eq!(world.remove_map.len(), 1);
         assert_eq!(world.remove_map[&(1, TypeId::of::<(i32,)>())], 2);
     }
 
+    #[test]
+    fn insert_can_be_used_to_overwrite_components() {
+        let drop1 = Rc::new(Cell::new(false));
+        let drop2 = Rc::new(Cell::new(false));
+        let drop3 = Rc::new(Cell::new(false));
+
+        let mut world = World::new();
+
+        let entity1 = world.alloc();
+        world.insert(entity1, (0, SetOnDrop(drop1.clone())));
+
+        let entity2 = world.alloc();
+
+        world.insert(entity1, (1, SetOnDrop(drop2.clone())));
+        world.insert(entity2, (2, SetOnDrop(drop3.clone())));
+
+        assert_eq!(*world.get::<i32>(entity1).unwrap(), 1);
+        assert_eq!(*world.get::<i32>(entity2).unwrap(), 2);
+
+        assert_eq!(world.insert_map.len(), 2);
+        assert_eq!(world.insert_map[&(0, TypeId::of::<(i32, SetOnDrop)>())], 1);
+        assert_eq!(world.insert_map[&(1, TypeId::of::<(i32, SetOnDrop)>())], 1);
+
+        assert!(drop1.get());
+        assert!(!drop2.get());
+        assert!(!drop3.get());
+    }
+
+    #[test]
+    fn exchange_can_be_used_to_overwrite_components() {
+        let drop1 = Rc::new(Cell::new(false));
+        let drop2 = Rc::new(Cell::new(false));
+
+        let mut world = World::new();
+
+        let entity = world.alloc();
+        world.insert(entity, (0, true, SetOnDrop(drop1.clone())));
+        world
+            .exchange::<(bool,), _>(entity, (1, SetOnDrop(drop2.clone())))
+            .unwrap();
+
+        assert_eq!(*world.get::<i32>(entity).unwrap(), 1);
+        assert!(!world.contains::<bool>(entity));
+
+        assert_eq!(world.exchange_map.len(), 1);
+        assert_eq!(
+            world.exchange_map[&(1, TypeId::of::<((bool,), (i32, SetOnDrop))>())],
+            2
+        );
+
+        assert!(drop1.get());
+        assert!(!drop2.get());
+    }
+
+    #[test]
+    fn empty_remove_is_essentially_a_noop() {
+        let mut world = World::new();
+
+        let ent = world.alloc();
+        world.insert(ent, (true,));
+
+        let () = world.remove::<()>(ent).unwrap();
+    }
+
     #[test]
     fn trival_exchange_does_not_create_aliasing_unique_references() {
         let mut world = World::new();
